﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Core;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Common;
using HeuristicLab.Common.Resources;
using HeuristicLab.Data;

namespace HeuristicLab.OptimizationExpertSystem {
  [Item("Expert-System", "Currently in experimental phase, an expert system that makes algorithm suggestions based on fitness landscape analysis features and an optimization knowledge base.")]
  [StorableClass]
  [Creatable(CreatableAttribute.Categories.TestingAndAnalysis, Priority = 119)]
  public sealed class ExpertSystem : NamedItem, IStorableContent, INotifyPropertyChanged {

    public string Filename { get; set; }

    public static new Image StaticItemImage {
      get { return VSImageLibrary.Library; }
    }

    [Storable]
    private int maximumEvaluations;
    public int MaximumEvaluations {
      get { return maximumEvaluations; }
      set {
        if (maximumEvaluations == value) return;
        maximumEvaluations = value;
        OnPropertyChanged("MaximumEvaluations");
        UpdateSuggestions();
      }
    }

    [Storable]
    private RunCollection runs;
    public RunCollection Runs {
      get { return runs; }
    }

    [Storable]
    private ItemList<IAlgorithm> algorithmInstances;
    public ItemList<IAlgorithm> AlgorithmInstances {
      get { return algorithmInstances; }
    }

    [Storable]
    private ISingleObjectiveHeuristicOptimizationProblem problem;
    public ISingleObjectiveHeuristicOptimizationProblem Problem {
      get { return problem; }
      set {
        if (problem == value) return;
        problem = value;
        OnPropertyChanged("Problem");
        UpdateSuggestions();
      }
    }

    [Storable]
    private ItemList<IAlgorithm> suggestedInstances;
    private ReadOnlyItemList<IAlgorithm> readOnlySuggestedInstances;
    public ReadOnlyItemList<IAlgorithm> SuggestedInstances {
      get { return readOnlySuggestedInstances; }
    }

    private bool Maximization {
      get { return Problem != null && ((IValueParameter<BoolValue>)Problem.MaximizationParameter).Value.Value; }
    }

    [StorableConstructor]
    private ExpertSystem(bool deserializing) : base(deserializing) { }
    private ExpertSystem(ExpertSystem original, Cloner cloner)
      : base(original, cloner) {
      runs = cloner.Clone(original.runs);
      algorithmInstances = cloner.Clone(original.algorithmInstances);
      problem = cloner.Clone(original.problem);
      suggestedInstances = cloner.Clone(original.suggestedInstances);
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      RegisterEventHandlers();
    }
    public ExpertSystem() {
      Name = ItemName;
      Description = ItemDescription;
      runs = new RunCollection();
      algorithmInstances = new ItemList<IAlgorithm>();
      suggestedInstances = new ItemList<IAlgorithm>();
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      RegisterEventHandlers();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ExpertSystem(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      RegisterEventHandlers();
    }

    private void RegisterEventHandlers() {
      runs.CollectionReset += InformationChanged;
      runs.ItemsAdded += InformationChanged;
      runs.ItemsRemoved += InformationChanged;
      runs.Reset += InformationChanged;
      runs.UpdateOfRunsInProgressChanged += InformationChanged;
      algorithmInstances.CollectionReset += InformationChanged;
      algorithmInstances.ItemsAdded += InformationChanged;
      algorithmInstances.ItemsRemoved += InformationChanged;
    }

    private void InformationChanged(object sender, EventArgs e) {
      var runCollection = sender as RunCollection;
      if (runCollection != null && runCollection.UpdateOfRunsInProgress) return;
      UpdateSuggestions();
    }

    private void UpdateSuggestions() {
      if (Problem == null) return;
      var instances = new SortedList<double, IAlgorithm>();
      foreach (var instance in algorithmInstances) {
        var relevantRuns = instance.Runs.Where(x => ((StringValue)x.Parameters["Problem Type"]).Value == Problem.GetType().Name);
        var avgQuality = 0.0;
        var counter = 0;
        foreach (var run in relevantRuns) {
          var performanceGraph = ((IndexedDataTable<double>)run.Results["QualityPerEvaluations"]);
          try {
            avgQuality += performanceGraph.Rows.First().Values.TakeWhile(x => x.Item1 < MaximumEvaluations).Last().Item2;
            counter++;
          } catch { continue; }
        }
        avgQuality /= counter;
        instances.Add(avgQuality, instance);
      }

      suggestedInstances.Clear();
      var instanceLadder = instances.Select(x => x.Value);
      suggestedInstances.AddRange(Maximization ? instanceLadder.Reverse() : instanceLadder);
    }

    public event PropertyChangedEventHandler PropertyChanged;
    private void OnPropertyChanged(string propertyName) {
      var handler = PropertyChanged;
      if (handler != null) handler(this, new PropertyChangedEventArgs(propertyName));
    }
  }
}
